from typing import Type, Set
from unittest import TestCase

from stereotype import Model, Role, ConfigurationError, DEFAULT_ROLE, IntField

ROLE_A = Role('a')
ROLE_B = Role('b', empty_by_default=True)
ROLE_C = Role('c')
ROLE_UNKNOWN_ALL = Role('unknown_1')
ROLE_ALL = Role('all')
ROLE_NONE = Role('none', empty_by_default=True)
ROLE_UNKNOWN_NONE = Role('unknown_2', empty_by_default=True)


class MyRoles(Model):
    a1: int = 1
    b1: float = 1.1
    a2: int = IntField(to_primitive_name='a_2', default=2)
    c1: str
    b2: float = 2.2
    hidden: int = IntField(to_primitive_name=None, default=404)

    @classmethod
    def declare_roles(cls):
        yield ROLE_A.whitelist(cls.a1, cls.a2)
        yield ROLE_B.blacklist(cls.a1, cls.a2, cls.c1)
        yield ROLE_C.whitelist(cls.c1)
        yield ROLE_ALL.blacklist()
        yield ROLE_NONE.whitelist()


class TestModels(TestCase):
    def test_roles_direct(self):
        model = MyRoles({'c1': '1'})
        self._assert_my_roles(model)

    def test_role_inheritance_simple(self):
        class MyChildRoles(MyRoles):
            pass

        self.assertEqual(['a1', 'b1', 'a_2', 'c1', 'b2'], MyChildRoles.field_names_for_role(DEFAULT_ROLE))

        model = MyChildRoles({'c1': 1})
        self._assert_my_roles(model)

        self.assertEqual(['a1', 'a_2'], MyChildRoles.field_names_for_role(ROLE_A))
        self.assertEqual(['b1', 'b2'], model.field_names_for_role(ROLE_B))
        self.assertEqual(['a1', 'b1', 'a_2', 'c1', 'b2'], MyChildRoles.field_names_for_role(ROLE_UNKNOWN_ALL))
        self.assertEqual([], MyChildRoles.field_names_for_role(ROLE_UNKNOWN_NONE))

    def _assert_my_roles(self, model: Model):
        all_serialized = {'a1': 1, 'a_2': 2, 'b1': 1.1, 'b2': 2.2, 'c1': '1'}
        self.assertEqual(all_serialized, model.serialize(role=DEFAULT_ROLE))
        self.assertEqual({'a1': 1, 'a_2': 2}, model.serialize(role=ROLE_A))
        self.assertEqual({'b1': 1.1, 'b2': 2.2}, model.serialize(role=ROLE_B))
        self.assertEqual({'c1': '1'}, model.to_primitive(role=ROLE_C))
        self.assertEqual(all_serialized, model.serialize(role=ROLE_UNKNOWN_ALL))
        self.assertEqual(all_serialized, model.serialize(role=ROLE_ALL))
        self.assertEqual({}, model.serialize(role=ROLE_NONE))
        self.assertEqual({}, model.serialize(role=ROLE_UNKNOWN_NONE))

    def test_role_inheritance(self):
        class MyOtherBase(Model):
            __abstract__ = True

            a3: int = 3

        class MyChildRoles(MyRoles, MyOtherBase):
            a4: int = 4
            b3: float = 3.3

            @classmethod
            def declare_roles(cls):
                yield ROLE_A.whitelist(cls.a4)
                yield ROLE_B.blacklist(cls.a3, cls.a4)

        model = MyChildRoles({'c1': 1})
        all_serialized = {'a1': 1, 'a_2': 2, 'b1': 1.1, 'b2': 2.2, 'c1': '1', 'a3': 3, 'a4': 4, 'b3': 3.3}
        self.assertEqual(all_serialized, model.serialize())
        self.assertEqual({'a1': 1, 'a_2': 2, 'a3': 3, 'a4': 4}, model.serialize(role=ROLE_A))
        self.assertEqual({'b1': 1.1, 'b2': 2.2, 'b3': 3.3}, model.serialize(role=ROLE_B))
        self.assertEqual({'a3': 3, 'a4': 4, 'b3': 3.3, 'c1': '1'}, model.serialize(role=ROLE_C))
        self.assertEqual(all_serialized, model.serialize(role=ROLE_UNKNOWN_ALL))
        self.assertEqual(all_serialized, model.serialize(role=ROLE_ALL))
        self.assertEqual({}, model.serialize(role=ROLE_NONE))
        self.assertEqual({}, model.serialize(role=ROLE_UNKNOWN_NONE))
        self.assertEqual("<MyChildRoles {a3=3, a1=1, b1=1.1, a2=2, c1='1', b2=2.2, hidden=404, a4=4, b3=3.3}>",
                         repr(model))

        # Abstract models can still be used if necessary, but don't have the __slots__ optimization, so it's not wise
        other = MyOtherBase({'a3': 3.0})
        self.assertEqual({'a3': 3}, other.serialize())
        self.assertEqual({'a3': 3}, other.serialize(role=ROLE_A))
        self.assertEqual({}, other.serialize(role=ROLE_B))
        self.assertEqual({'a3': 3}, other.serialize(role=ROLE_C))
        self.assertEqual({'a3': 3}, other.serialize(role=ROLE_UNKNOWN_ALL))
        self.assertEqual({'a3': 3}, other.serialize(role=ROLE_ALL))
        self.assertEqual({}, other.serialize(role=ROLE_NONE))
        self.assertEqual({}, other.serialize(role=ROLE_UNKNOWN_NONE))

    def test_role_inheritance_with_override(self):
        class MyChildRoles(MyRoles):
            a3: int = 3
            b3: float = 3.3

            @classmethod
            def declare_roles(cls):
                yield ROLE_A.whitelist(cls.a2, cls.a3, override_parents=True)
                yield ROLE_B.blacklist(cls.a1, cls.a3, override_parents=True)
                yield ROLE_C.whitelist(cls.c1, override_parents=True)
                yield ROLE_ALL.whitelist(override_parents=True)
                yield ROLE_NONE.blacklist(override_parents=True)

        model = MyChildRoles({'c1': 1})
        all_serialized = {'a1': 1, 'a_2': 2, 'a3': 3, 'b1': 1.1, 'b2': 2.2, 'b3': 3.3, 'c1': '1'}
        self.assertEqual(all_serialized, model.serialize())
        self.assertEqual({'a_2': 2, 'a3': 3}, model.serialize(role=ROLE_A))
        self.assertEqual({'a_2': 2, 'b1': 1.1, 'b2': 2.2, 'b3': 3.3, 'c1': '1'}, model.serialize(role=ROLE_B))
        self.assertEqual({'c1': '1'}, model.serialize(role=ROLE_C))
        self.assertEqual(all_serialized, model.serialize(role=ROLE_UNKNOWN_ALL))
        self.assertEqual({}, model.serialize(role=ROLE_ALL))
        self.assertEqual(all_serialized, model.serialize(role=ROLE_NONE))
        self.assertEqual({}, model.serialize(role=ROLE_UNKNOWN_NONE))

    def test_nested_model_roles(self):
        class ChildA(Model):
            a: str = 'default'

            @classmethod
            def declare_roles(cls):
                yield ROLE_C.blacklist(cls.a)

        class ChildB(ChildA):
            b: int = 7

            @classmethod
            def declare_roles(cls):
                yield ROLE_B.whitelist(cls.b)

        class Parent(Model):
            a: ChildA = ChildA
            b: ChildB = ChildB

            @classmethod
            def declare_roles(cls):
                yield ROLE_B.blacklist(cls.a)

            @classmethod
            def resolve_extra_types(cls) -> Set[Type[Model]]:
                return {ChildA, ChildB}

        model = Parent()
        self.assertEqual({'a': {'a': 'default'}, 'b': {'a': 'default', 'b': 7}}, model.serialize())
        self.assertEqual({'a': {'a': 'default'}, 'b': {'a': 'default', 'b': 7}}, model.serialize(role=ROLE_A))
        self.assertEqual({'b': {'b': 7}}, model.serialize(role=ROLE_B))
        self.assertEqual({'a': {}, 'b': {'b': 7}}, model.serialize(role=ROLE_C))
        self.assertEqual({'a': {'a': 'default'}, 'b': {'a': 'default', 'b': 7}}, model.serialize(role=ROLE_ALL))
        self.assertEqual({}, model.serialize(role=ROLE_NONE))

    def test_role_configuration_error_conflict(self):
        class BadRoles(Model):
            whatever: int = 0

            @classmethod
            def declare_roles(cls):
                yield ROLE_A.whitelist(cls.whatever)
                yield ROLE_B.blacklist(cls.whatever)
                yield ROLE_A.whitelist()

        with self.assertRaises(ConfigurationError) as ctx:
            BadRoles({'whatever': 1})
        self.assertEqual('Role a configured for BadRoles multiple times', str(ctx.exception))

    def test_role_configuration_error_bad_field(self):
        class BadRoles(Model):
            not_a_field: int = 0

            @classmethod
            def declare_roles(cls):
                yield ROLE_A.whitelist('not_a_field')

        with self.assertRaises(ConfigurationError) as ctx:
            BadRoles({'not_a_field': 1})
        self.assertEqual("Role blacklist/whitelist needs member descriptors (e.g. cls.my_field), got 'not_a_field'",
                         str(ctx.exception))

    def test_role_repr(self):
        self.assertEqual('<Role default, empty_by_default=False, code=0>', repr(DEFAULT_ROLE))
        self.assertEqual('<Role a, empty_by_default=False, code=1>', repr(ROLE_A))
        self.assertEqual('<Role b, empty_by_default=True, code=2>', repr(ROLE_B))



import gc
import json
from time import sleep, time
from typing import Callable, List, Iterable


def make_stuff_data(items: int):
    return {
        'name': 'xyz',
        'flag': True,
        'model': {'value': 42},
        'items': [{
            'name': 'abc',
            'stuff': make_stuff_data(items - 1),
            'list': [str(i) for i in range(items)],
        } for _ in range(items)],
    }


def benchmark_cpu(executable: Callable[[List[dict], bool], Iterable[dict]], depth: int = 2, validate: bool = False):
    # Execute once just to initialize the classes
    list(executable([make_stuff_data(depth)], validate))

    # Execute 10 times to get an initial estimate to make the run time 1 second each INCLUDING making the data
    start = float(time())
    data = [make_stuff_data(depth) for _ in range(10)]
    for _ in executable(data, validate):
        pass
    print(time(), start)
    estimate = (time() - start) / 10

    n = round(1.0 / estimate)
    print(f'Iterations: {n}, time reported in milliseconds:')

    best = float('inf')
    for i in range(10):
        data = [make_stuff_data(depth) for _ in range(n)]
        start = time()
        for _ in executable(data, validate):
            pass
        milliseconds = 1e3 * (time() - start) / n
        best = min(best, milliseconds)
        print(f'Run {i}: {milliseconds:.3f}')

    print(f'Ideal: {best:.3f}')


def execute(model, iterations, depth):
    stuff = model(make_stuff_data(1))
    print(stuff)
    try:
        stuff.xyz = 'Slots restricted: NO'
        print(stuff.xyz)
    except (AttributeError, ValueError):
        print('Slots restricted: YES')
    stuff_serialized = json.dumps(stuff.serialize())
    #print(len(stuff_serialized))
    #print(stuff_serialized)

    serialized = make_stuff_data(depth)
    data = [make_stuff_data(depth) for i in range(iterations)]
    objects = [None] * iterations
    started = time()
    for i in range(iterations):
        stuff = model(data[i])
        serialized = stuff.serialize()
        #stuff.validate()
        objects[i] = stuff
        data[i] = None
    ended = time()
    print(f'{(ended - started) / max(iterations, 1):0.9f}')
    #print(len(json.dumps(serialized)))
    #print(type(objects))


def memory_bench(model, serialize, iterations, depth):
    json_data = json.dumps(serialize(model(make_stuff_data(depth))))
    print(len(json_data))
    started = time()
    # initial = make_stuff_data(depth)
    objects = [model(make_stuff_data(depth)) for i in range(iterations)]
    ended = time()
    print(f'{(ended - started) / max(iterations, 1):0.9f}')
    print(type(objects))
    gc.collect()
    sleep(10)
    del objects
    gc.collect()
    sleep(10)



from __future__ import annotations

from typing import Union, Any, Iterable, Optional

from stereotype.fields.annotations import AnnotationResolver
from stereotype.fields.base import Field
from stereotype.utils import Missing, ConfigurationError, PathErrorType, ValidationContextType


class _AtomicField(Field):
    atomic = True

    def init_from_annotation(self, parser: AnnotationResolver):
        if parser.annotation is not self.type:
            raise parser.incorrect_type(self)


class BoolField(_AtomicField):
    __slots__ = Field.__slots__
    type = bool
    type_repr = 'bool'
    empty_value = False

    def __init__(self, *, default: Any = Missing, hide_none: bool = False, hide_false: bool = False,
                 primitive_name: Optional[str] = Missing, to_primitive_name: Optional[str] = Missing):
        super().__init__(default=default, hide_none=hide_none, hide_empty=hide_false,
                         primitive_name=primitive_name, to_primitive_name=to_primitive_name)

    def convert(self, value: Any) -> Any:
        if value is Missing:
            return self._fill_missing()
        if value is None:
            return None
        if value is False or value is True:
            return value
        if value in {'true', 'True', 'yes', 'Yes'}:
            return True
        if value in {'false', 'False', 'no', 'No'}:
            return False
        raise TypeError('Value must be a boolean or a true/false/yes/no string value')


class _BaseNumberField(_AtomicField):
    __slots__ = Field.__slots__ + ('min_value', 'max_value')
    min_value: Union[int, float, None]
    max_value: Union[int, float, None]

    def _set_min_max_value_validation(self, min_value: Union[int, float, None], max_value: Union[int, float, None]):
        if min_value is not None and max_value is not None:
            self.native_validate = self._validate_min_max_value
        elif min_value is not None:
            self.native_validate = self._validate_min_value
        elif max_value is not None:
            self.native_validate = self._validate_max_value

    def _validate_min_max_value(self, value: Any, _: ValidationContextType) -> Iterable[PathErrorType]:
        if not (self.min_value <= value <= self.max_value):
            yield (), f'Must be between {self.min_value} and {self.max_value}'

    def _validate_min_value(self, value: Any, _: ValidationContextType) -> Iterable[PathErrorType]:
        if value < self.min_value:
            yield (), f'Must be at least {self.min_value}'

    def _validate_max_value(self, value: Any, _: ValidationContextType) -> Iterable[PathErrorType]:
        if value > self.max_value:
            yield (), f'Must be at most {self.max_value}'


class IntField(_BaseNumberField):
    __slots__ = _BaseNumberField.__slots__
    type = int
    type_repr = 'int'
    empty_value = 0

    def __init__(self, *, default: Any = Missing, hide_none: bool = False, hide_zero: bool = False,
                 primitive_name: Optional[str] = Missing, to_primitive_name: Optional[str] = Missing,
                 min_value: Optional[int] = None, max_value: Optional[int] = None):
        super().__init__(default=default, hide_none=hide_none, hide_empty=hide_zero,
                         primitive_name=primitive_name, to_primitive_name=to_primitive_name)
        self.min_value = min_value
        self.max_value = max_value
        self._set_min_max_value_validation(min_value, max_value)

    def convert(self, value: Any) -> Any:
        if value is Missing:
            return self._fill_missing()
        if value is None:
            return None
        if isinstance(value, float) and value != int(value):
            raise TypeError(f'Numeric value {value} is not an integer')
        try:
            return int(value)
        except (ValueError, TypeError):
            raise TypeError(f'Value {value!r} is not an integer number')


class FloatField(_BaseNumberField):
    __slots__ = _BaseNumberField.__slots__
    type = float
    type_repr = 'float'
    empty_value = 0.

    def __init__(self, *, default: Any = Missing, hide_none: bool = False, hide_zero: bool = False,
                 primitive_name: Optional[str] = Missing, to_primitive_name: Optional[str] = Missing,
                 min_value: Optional[float] = None, max_value: Optional[float] = None):
        super().__init__(default=default, hide_none=hide_none, hide_empty=hide_zero,
                         primitive_name=primitive_name, to_primitive_name=to_primitive_name)
        self.min_value = min_value
        self.max_value = max_value
        self._set_min_max_value_validation(min_value, max_value)


class StrField(_AtomicField):
    __slots__ = _AtomicField.__slots__ + ('min_length', 'max_length', 'choices')
    type = str
    type_repr = 'str'
    empty_value = ''

    def __init__(self, *, default: Any = Missing, hide_none: bool = False, hide_empty: bool = False,
                 primitive_name: Optional[str] = Missing, to_primitive_name: Optional[str] = Missing,
                 min_length: int = 0, max_length: Optional[int] = None, choices: Optional[Iterable[str]] = None):
        super().__init__(default=default, hide_none=hide_none, hide_empty=hide_empty,
                         primitive_name=primitive_name, to_primitive_name=to_primitive_name)
        if (min_length > 0 or max_length is not None) and choices is not None:
            raise ConfigurationError('Cannot use min_length or max_length together with choices')
        self.min_length = min_length
        self.max_length = max_length
        self.choices = {choice: None for choice in choices} if choices is not None else None  # Sets are not ordered
        if self.choices is not None:
            self.native_validate = self._validate_choices
        elif min_length > 0 and max_length is not None:
            self.native_validate = self._validate_min_max_length
        elif min_length == 1:
            self.native_validate = self._validate_not_empty
        elif min_length > 0:
            self.native_validate = self._validate_min_length
        elif max_length is not None:
            self.native_validate = self._validate_max_length

    def _validate_choices(self, value: str, _: ValidationContextType) -> Iterable[PathErrorType]:
        if value not in self.choices:
            yield (), f'Must be one of: {", ".join(self.choices)}'

    def _validate_min_max_length(self, value: str, _: ValidationContextType) -> Iterable[PathErrorType]:
        if not (self.min_length <= len(value) <= self.max_length):
            if self.min_length == self.max_length:
                yield (), f'Must be exactly {self.min_length} character{"s" if self.min_length > 1 else ""} long'
            else:
                yield (), f'Must be {self.min_length} to {self.max_length} characters long'

    def _validate_not_empty(self, value: str, _: ValidationContextType) -> Iterable[PathErrorType]:
        if not value:
            yield (), 'This value cannot be empty'

    def _validate_min_length(self, value: str, _: ValidationContextType) -> Iterable[PathErrorType]:
        if len(value) < self.min_length:
            yield (), f'Must be at least {self.min_length} characters long'

    def _validate_max_length(self, value: str, _: ValidationContextType) -> Iterable[PathErrorType]:
        if len(value) > self.max_length:
            yield (), f'Must be at most {self.max_length} character{"s" if self.max_length > 1 else ""} long'


ATOMIC_TYPE_MAPPING = {
    bool: BoolField,
    int: IntField,
    float: FloatField,
    str: StrField,
}



from __future__ import annotations

from typing import Optional, Iterable
from unittest import TestCase

from schematics.models import Model as SchematicsModel
from schematics.transforms import blacklist, whitelist
from schematics.types import IntType, StringType, FloatType, BooleanType
from schematics.types.compound import ModelType, ListType

from stereotype import Model, StrField, Missing, Role, ValidationError, RequestedRoleFields, ConfigurationError
from stereotype.compat.schematics import SchematicsModelField


ROLE_X = Role('x')
ROLE_Y = Role('y')


class Inner(SchematicsModel):
    stuff = ListType(FloatType(required=True), required=True, default=list, max_size=2)
    bool = BooleanType(required=True)

    class Options:
        roles = {
            ROLE_X.name: blacklist('stuff'),
            ROLE_Y.name: whitelist('stuff'),
        }


class Root(Model):
    number: float = 42.
    string: str = StrField(min_length=3)
    inner: Inner = SchematicsModelField(default=Inner)
    outer: Optional[Outer] = SchematicsModelField(hide_none=True, primitive_name='otter')

    @classmethod
    def declare_roles(cls) -> Iterable[RequestedRoleFields]:
        yield ROLE_X.blacklist(cls.number)


class Outer(SchematicsModel):
    integer = IntType(required=True)
    string = StringType(required=False, default="def", choices=['str', 'ing'])
    recursive = ModelType(Inner)

    @staticmethod
    def validate_integer(value, _):
        if value % 2:
            raise ValueError('Must be even')

    class Options:
        roles = {
            ROLE_X.name: blacklist(),
            ROLE_Y.name: whitelist('integer', 'recursive'),
        }


class TestSchematicsModelField(TestCase):
    def test_empty(self):
        root = Root()
        self.assertEqual(Inner(), root.inner)
        self.assertIs(Missing, root.outer)
        self.assertEqual({'inner': {'bool': None, 'stuff': []}, 'number': 42.0}, root.serialize())
        self.assertEqual(root.copy(deep=True), root)
        self.assertEqual('<Root {number=42.0, string=Missing, inner=<Inner: Inner object>, outer=Missing}>', repr(root))

    def test_none_value(self):
        root = Root({'inner': None, 'otter': None})
        self.assertIsNone(root.inner)
        self.assertIsNone(root.outer)
        self.assertEqual({'inner': None, 'number': 42.0}, root.serialize())
        self.assertEqual(root.copy(deep=True), root)
        with self.assertRaises(ValidationError) as e:
            root.validate()
        self.assertEqual({
            'inner': ['This field is required'],
            'string': ['This field is required'],
        }, e.exception.errors)

    def test_full(self):
        root = Root({
            'number': 47,
            'string': False,
            'inner': {'stuff': [1, 2, 3], 'bool': 1},
            'otter': {'integer': '7', 'string': 1, 'recursive': {'stuff': [1., 2., 3.]}},
        })
        self.assertEqual(Inner({'stuff': [1, 2, 3], 'bool': True}), root.inner)
        self.assertEqual('1', root.outer.string)
        self.assertEqual({
            'number': 47.,
            'string': 'False',
            'inner': {'stuff': [1., 2., 3.], 'bool': True},
            'otter': {'integer': 7, 'string': '1', 'recursive': {'stuff': [1., 2., 3.], 'bool': None}},
        }, root.to_primitive())

        copied = root.copy(deep=True)
        self.assertEqual(root, copied)
        copied.string = "True"
        self.assertNotEqual(root, copied)
        self.assertEqual({
            'string': 'True',
            'inner': {'bool': True},
            'otter': {'integer': 7, 'string': '1', 'recursive': {'bool': None}},
        }, copied.serialize(role=ROLE_X))
        self.assertEqual({
            'number': 47.,
            'string': 'True',
            'inner': {'stuff': [1., 2., 3.]},
            'otter': {'integer': 7, 'recursive': {'stuff': [1., 2., 3.]}},
        }, copied.to_primitive(role=ROLE_Y))

        with self.assertRaises(ValidationError) as e:
            root.validate()
        self.assertEqual({
            'inner': {'stuff': ['Please provide no more than 2 items.']},
            'otter': {
                'recursive': {'bool': ['This field is required.'], 'stuff': ['Please provide no more than 2 items.']},
                'string': ["Value must be one of ['str', 'ing']."],
            },
        }, e.exception.errors)

    def test_configuration_error_no_explicit_field(self):
        class Bad(Model):
            worse: Outer = Outer
        with self.assertRaises(ConfigurationError) as e:
            Bad()
        self.assertEqual("Field worse: Unrecognized field annotation Outer (may need an explicit Field)",
                         str(e.exception))

    def test_configuration_error_mismatch(self):
        class Bad(Model):
            worse: Root = SchematicsModelField()
        with self.assertRaises(ConfigurationError) as e:
            Bad()
        self.assertEqual("Field worse: SchematicsModelField cannot be used for annotation Root, should use ModelField",
                         str(e.exception))



from __future__ import annotations

from typing import Optional, Any, Union, List

from stereotype import Model


class Variable(Model):
    name: str


class IsEqual(Model):
    type = 'equal'

    variable: Variable
    constant: Constant


class Constant(Model):
    value: Optional[Any] = None


equality = IsEqual({'variable': {'name': 'x'}, 'constant': {'value': 42}})
assert equality.variable.name == 'x'
assert equality.type == 'equal'
assert equality.to_primitive() == {'variable': {'name': 'x'}, 'constant': {'value': 42}}


class Negation(Model):
    type = 'not'

    formula: Union[IsEqual, Conjunction]


class Conjunction(Model):
    type = 'and'

    operands: List[Union[IsEqual, Negation, Conjunction]] = list


equality_copy = equality.copy(deep=True)
model = Conjunction({'operands': [
    {'type': 'not', 'formula': equality_copy},
    {'type': 'equal', 'variable': {'name': 'y'}, 'constant': {}},
]})
assert isinstance(model.operands[0], Negation)
assert model.operands[0].type == 'not'
assert model.to_primitive() == {'operands': [
    {'type': 'not', 'formula': {'type': 'equal', 'variable': {'name': 'x'}, 'constant': {'value': 42}}},
    {'type': 'equal', 'variable': {'name': 'y'}, 'constant': {'value': None}},
]}



from __future__ import annotations

from typing import Any, Optional, Type, Iterable, Tuple, Dict, cast, Union, get_args

from stereotype.fields.annotations import AnnotationResolver
from stereotype.fields.base import Field
from stereotype.model import Model
from stereotype.roles import Role, DEFAULT_ROLE
from stereotype.utils import Missing, ConfigurationError, PathErrorType, ValidationContextType


class ModelField(Field):
    __slots__ = Field.__slots__ + ('type',)
    atomic = False
    empty_value = {}

    def __init__(self, *, default: Any = Missing, hide_none: bool = False, hide_empty: bool = False,
                 primitive_name: Optional[str] = Missing, to_primitive_name: Optional[str] = Missing):
        super().__init__(default=default, hide_none=hide_none, hide_empty=hide_empty,
                         primitive_name=primitive_name, to_primitive_name=to_primitive_name)
        self.type: Type[Model] = cast(Type[Model], NotImplemented)
        self.native_validate = self.validate

    def init_from_annotation(self, parser: AnnotationResolver):
        if not issubclass(parser.annotation, Model):
            raise parser.incorrect_type(self)
        self.type = parser.annotation

    def validate(self, value: Model, context: ValidationContextType) -> Iterable[PathErrorType]:
        yield from value.validation_errors(context)

    def convert(self, value: Any) -> Any:
        if value is Missing:
            if self.required:
                return Missing
            return self.default if self.default_factory is None else self.default_factory()
        if value is None:
            return None
        if isinstance(value, self.type):
            return value
        if not isinstance(value, dict):
            raise TypeError(f'Supplied type {type(value).__name__}, needs a mapping or {self.type.__name__}')
        return self.type(value)

    def copy_value(self, value: Any) -> Any:
        if value is None or value is Missing:
            return value
        return value.copy(deep=True)

    def to_primitive(self, value: Any, role: Role = DEFAULT_ROLE) -> Any:
        if value is None or value is Missing:
            return value
        return value.to_primitive(role)

    @property
    def type_repr(self):
        return self.type.__name__ if self.type is not NotImplemented else '(unknown)'


class DynamicModelField(Field):
    __slots__ = Field.__slots__ + ('types', 'type_map')
    atomic = False
    type = Model

    def __init__(self, *, default: Any = Missing, hide_none: bool = False,
                 primitive_name: Optional[str] = Missing, to_primitive_name: Optional[str] = Missing):
        super().__init__(default=default, hide_none=hide_none,
                         primitive_name=primitive_name, to_primitive_name=to_primitive_name)
        self.types: Tuple[Type[Model], ...] = NotImplemented
        self.type_map: Dict[str, Type[Model]] = NotImplemented
        self.native_validate = self.validate

    def init_from_annotation(self, parser: AnnotationResolver):
        if parser.origin is not Union:
            raise parser.incorrect_type(self)
        options = get_args(parser.annotation)

        if not all(issubclass(option, Model) for option in options):
            raise ConfigurationError(f'Union Model fields can only be Optional or Union of Model subclass types, '
                                     f'got {parser!r}')

        type_map = {}
        for option in options:
            if not hasattr(option, 'type'):
                raise ConfigurationError(f"Model {option.__name__} used in a dynamic model field {parser!r} but "
                                         f"does not define a non-type-annotated string `type` field")
            if type(option.type).__name__ == 'member_descriptor':
                raise ConfigurationError(f"Model {option.__name__} used in a dynamic model field {parser!r} but its "
                                         f"`type` field has a type annotation making it a field, must be an attribute")
            if not isinstance(option.type, str):
                raise ConfigurationError(f"Model {option.__name__} used in a dynamic model field {parser!r} but its "
                                         f"`type` field {option.type} is not a string")
            if option.type in type_map:
                raise ConfigurationError(f"Conflicting dynamic model field types in {parser!r}: "
                                         f"{type_map[option.type].__name__} vs {option.__name__}")
            type_map[option.type] = option

        self.type_map = type_map
        self.types = tuple(type_map.values())

    def validate(self, value: Any, context: ValidationContextType) -> Iterable[PathErrorType]:
        yield from value.validation_errors(context)

    def convert(self, value: Any) -> Any:
        if value is Missing:
            if self.required:
                return Missing
            return self.default if self.default_factory is None else self.default_factory()
        is_model = isinstance(value, Model)
        if is_model and not isinstance(value, self.types):
            raise TypeError(f'Expected {self.type_repr}, got {type(value).__name__}')
        if is_model or value is None:
            return value

        try:
            value_type = value['type']
        except TypeError:
            raise TypeError(f'Expected a mapping with a `type` field, got type {type(value).__name__}')
        except KeyError:
            raise TypeError('Expected a mapping with a `type` field, got no `type` field')
        type_cls = self.type_map.get(value_type, None)
        if type_cls is None:
            raise TypeError(f'Got a mapping with unsupported `type` {value_type!r}')
        return type_cls(value)

    def copy_value(self, value: Any) -> Any:
        if value is None or value is Missing:
            return value
        return value.copy(deep=True)

    def to_primitive(self, value: Any, role: Role = DEFAULT_ROLE) -> Any:
        if value is None or value is Missing:
            return value
        result = value.to_primitive(role)
        result['type'] = value.type
        return result

    @property
    def type_repr(self):
        return f'Union[{", ".join(option.__name__ for option in self.types)}]'



from typing import List

from schematics.exceptions import ValidationError
from schematics.models import Model
from schematics.types import StringType, BooleanType, IntType, FloatType
from schematics.types.compound import ListType, ModelType

from bench.common import memory_bench, benchmark_cpu


class BaseStuff(Model):
    name = StringType(required=True, min_length=2)
    flag = BooleanType(required=True, default=False)


class ListStuff(BaseStuff):
    stuff = ModelType('bench.bench_schematics2.Stuff', required=True)
    list = ListType(StringType(required=True), required=True, min_size=1)


class ModelStuff(Model):
    value = FloatType(required=True)
    def_value = FloatType(required=False, default=4.2)


class Stuff(BaseStuff):
    optional = IntType(required=False, default=None)
    strange = FloatType(required=False, default=4.7)
    model = ModelType(ModelStuff, required=True)
    items = ListType(ModelType(ListStuff, required=True), required=True, default=list)

    def validate_flag(self, data, value):
        if not value and data['optional'] is None:
            raise ValidationError('Must be true if optional is not set')


if __name__ == '__main__':
    def benchmark(inputs: List[dict], validate: bool):
        for data in inputs:
            model = Stuff(data)
            if validate:
                model.validate()
            yield model.serialize()

    benchmark_cpu(benchmark, depth=4, validate=False)
    # memory_bench(Stuff, 10000, 2)



from __future__ import annotations

from typing import List, Optional

from stereotype import Model, StrField, ListField
from bench.common import benchmark_cpu, memory_bench


class BaseStuff(Model):
    name: str = StrField(min_length=2)
    flag: bool = False


class ListStuff(BaseStuff):
    stuff: Stuff
    list: List[str] = ListField(min_length=1)


class ModelStuff(Model):
    value: float
    def_value: float = 4.2


class Stuff(BaseStuff):
    optional: Optional[int] = None
    strange: Optional[float] = 4.7
    model: ModelStuff
    items: List[ListStuff] = list

    def validate_flag(self, value: bool, _):
        if not value and self.optional is None:
            raise ValueError('Must be true if optional is not set')


if __name__ == '__main__':
    def benchmark(inputs: List[dict], validate: bool):
        for data in inputs:
            model = Stuff(data)
            if validate:
                model.validate()
            yield model.to_primitive()

    benchmark_cpu(benchmark, depth=4, validate=False)
    # memory_bench(Stuff, lambda model: model.to_primitive(), 10000, 4)  # 349MB



from __future__ import annotations

from typing import Any, Optional, Iterable, get_args

from stereotype.fields.annotations import AnnotationResolver
from stereotype.fields.base import Field, ValidationContextType
from stereotype.roles import Role, DEFAULT_ROLE
from stereotype.utils import Missing, ConfigurationError, ConversionError, PathErrorType


class _CompoundField(Field):
    __slots__ = Field.__slots__ + ('min_length', 'max_length')
    atomic = False

    def __init__(self, *, default: Any = Missing, hide_none: bool = False, hide_empty: bool = False,
                 primitive_name: Optional[str] = Missing, to_primitive_name: Optional[str] = Missing,
                 min_length: int = 0, max_length: Optional[int] = None):
        super().__init__(default=default, hide_none=hide_none, hide_empty=hide_empty,
                         primitive_name=primitive_name, to_primitive_name=to_primitive_name)
        self.min_length = min_length
        self.max_length = max_length

    def init_from_annotation(self, parser: AnnotationResolver):
        raise NotImplementedError  # pragma: no cover

    def validate(self, value: Any, context: ValidationContextType) -> Iterable[PathErrorType]:
        if self.min_length > 0:
            if self.max_length is not None:
                if not (self.min_length <= len(value) <= self.max_length):
                    if self.min_length == self.max_length:
                        yield (), f'Provide exactly {self.min_length} item{"s" if self.min_length > 1 else ""}'
                    else:
                        yield (), f'Provide {self.min_length} to {self.max_length} items'
            elif len(value) < self.min_length:
                yield (), f'Provide at least {self.min_length} item{"s" if self.min_length > 1 else ""}'
        elif self.max_length is not None and len(value) > self.max_length:
            yield (), f'Provide at most {self.max_length} item{"s" if self.max_length > 1 else ""}'


class ListField(_CompoundField):
    __slots__ = _CompoundField.__slots__ + ('item_field',)
    type = list
    empty_value = []

    def __init__(self, item_field: Field = NotImplemented, *,
                 default: Any = Missing, hide_none: bool = False, hide_empty: bool = False,
                 primitive_name: Optional[str] = Missing, to_primitive_name: Optional[str] = Missing,
                 min_length: int = 0, max_length: Optional[int] = None):
        super().__init__(default=default, hide_none=hide_none, hide_empty=hide_empty,
                         primitive_name=primitive_name, to_primitive_name=to_primitive_name,
                         min_length=min_length, max_length=max_length)
        self.item_field: Field = item_field
        self.native_validate = self.validate

    def init_from_annotation(self, parser: AnnotationResolver):
        if parser.origin is not list:
            raise parser.incorrect_type(self)
        item_annotation, = get_args(parser.annotation)
        self.item_field = AnnotationResolver(item_annotation).resolve(self.item_field)

    def validate(self, value: Any, context: ValidationContextType) -> Iterable[PathErrorType]:
        yield from super().validate(value, context)
        item_field = self.item_field
        for index, item in enumerate(value):
            if item is Missing or (item is None and not item_field.allow_none):
                yield (str(index),), 'This field is required'
            elif item is not None and item_field.native_validate is not None:
                for path, error in item_field.native_validate(item, context):
                    yield (str(index),) + path, error

    def convert(self, value: Any) -> Any:
        if value is Missing:
            return self._fill_missing()
        if value is None:
            return None
        converter = self.item_field.convert
        converted = []
        error_index = 0
        try:
            for error_index, item in enumerate(value):
                converted.append(converter(item))
        except ConversionError as e:
            raise e.wrapped(str(error_index))
        except (TypeError, ValueError) as e:
            raise ConversionError.new(str(e), str(error_index))
        return converted

    def copy_value(self, value: Any) -> Any:
        if value is Missing or value is None:
            return value
        if self.item_field.atomic:
            return list(value)
        item_copy = self.item_field.copy_value
        return [item_copy(item) for item in value]

    def to_primitive(self, value: Any, role: Role = DEFAULT_ROLE) -> Any:
        if value is None or value is Missing:
            return value
        if self.item_field.atomic:
            return list(value)
        item_to_primitive = self.item_field.to_primitive
        return [item_to_primitive(item, role) for item in value]

    @property
    def type_repr(self):
        return f'List[{"?" if self.item_field is NotImplemented else self.item_field.type_repr}]'


class DictField(_CompoundField):
    __slots__ = _CompoundField.__slots__ + ('key_field', 'value_field')
    type = dict
    empty_value = {}

    def __init__(self, key_field: Field = NotImplemented, value_field: Field = NotImplemented, *,
                 default: Any = Missing, hide_none: bool = False, hide_empty: bool = False,
                 primitive_name: Optional[str] = Missing, to_primitive_name: Optional[str] = Missing,
                 min_length: int = 0, max_length: Optional[int] = None):
        super().__init__(default=default, hide_none=hide_none, hide_empty=hide_empty,
                         primitive_name=primitive_name, to_primitive_name=to_primitive_name,
                         min_length=min_length, max_length=max_length)
        self.key_field: Field = key_field
        self.value_field: Field = value_field
        self.native_validate = self.validate

    def init_from_annotation(self, parser: AnnotationResolver):
        if parser.origin is not dict:
            raise parser.incorrect_type(self)
        key_annotation, value_annotation = get_args(parser.annotation)
        self.key_field = AnnotationResolver(key_annotation).resolve(self.key_field)
        if not self.key_field.atomic:
            raise ConfigurationError(f'DictField keys may only be booleans, numbers or strings: {parser!r}')
        self.value_field = AnnotationResolver(value_annotation).resolve(self.value_field)

    def validate(self, value: Any, context: ValidationContextType) -> Iterable[PathErrorType]:
        yield from super().validate(value, context)
        key_field, value_field = self.key_field, self.value_field
        for key, val in value.items():
            reported_required = False
            if key is Missing or (key is None and not key_field.allow_none):
                reported_required = True
                yield (str(key),), 'This field is required'
            elif key is not None and key_field.native_validate is not None:
                for path, error in key_field.native_validate(key, context):
                    yield (str(key),) + path, error
            if val is Missing or (val is None and not value_field.allow_none):
                if not reported_required:
                    yield (str(key),), 'This field is required'
            elif val is not None and value_field.native_validate is not None:
                for path, error in value_field.native_validate(val, context):
                    yield (str(key),) + path, error

    def convert(self, value: Any) -> Any:
        if value is Missing:
            return self._fill_missing()
        if value is None:
            return None
        if not isinstance(value, dict):
            raise TypeError(f'Expected a dict, got a {type(value).__name__}')
        key_converter = self.key_field.convert
        value_converter = self.value_field.convert
        error_key = Missing  # An error cannot occur before the first assignment to this, so Missing won't be used
        try:
            return {key_converter(error_key := key): value_converter(val) for key, val in value.items()}
        except ConversionError as e:
            raise e.wrapped(str(error_key))
        except (TypeError, ValueError) as e:
            raise ConversionError.new(str(e), str(error_key))

    def copy_value(self, value: Any) -> Any:
        if value is None or value is Missing:
            return value
        if self.value_field.atomic:
            return dict(value)
        item_to_primitive = self.value_field.copy_value
        return {key: item_to_primitive(val) for key, val in value.items()}

    def to_primitive(self, value: Any, role: Role = DEFAULT_ROLE) -> Any:
        if value is None or value is Missing:
            return value
        if self.value_field.atomic:
            return dict(value)
        item_to_primitive = self.value_field.to_primitive
        return {key: item_to_primitive(val, role) for key, val in value.items()}

    @property
    def type_repr(self):
        key_repr = '?' if self.key_field is NotImplemented else self.key_field.type_repr
        value_repr = '?' if self.value_field is NotImplemented else self.value_field.type_repr
        return f'Dict[{key_repr}, {value_repr}]'



from __future__ import annotations

from typing import Optional, Union, List, Dict, cast, Iterable
from unittest import TestCase

from stereotype import Model, Missing, ValidationError, ConversionError, ModelField, ConfigurationError, ListField, \
    StrField, BoolField, FloatField
from stereotype.fields.compound import DictField
from stereotype.roles import RequestedRoleFields, Role


class MyBoolModel(Model):
    type = 'bool'
    field: bool
    bad_field: Optional[bool] = BoolField(default=None, hide_none=True)

    def validate_bad_field(self, value: Optional[bool], _):
        if value is True:
            raise ValueError('Nope')


class MyStrModel(Model):
    type = 'str'
    field: str = StrField(default='', max_length=3)


class SomeLists(Model):
    ints: List[int]
    optionals: List[Optional[str]] = ListField(StrField(min_length=2))
    lists: List[List[float]] = list
    models: Optional[List[MyBoolModel]] = None
    dynamics: List[Union[MyBoolModel, MyStrModel]] = ListField(primitive_name='union')


class SimpleInt(Model):
    int: int


class TestListType(TestCase):
    def test_empty(self):
        model = SomeLists()
        self.assertIs(Missing, model.ints)
        self.assertIs(Missing, model.optionals)
        self.assertEqual([], model.lists)
        self.assertIsNone(model.models)
        self.assertIs(Missing, model.dynamics)
        self.assertEqual('<SomeLists {ints=Missing, optionals=Missing, lists=[], models=None, dynamics=Missing}>',
                         repr(model))
        self.assertEqual({'lists': [], 'models': None}, model.serialize())
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual({
            'ints': ['This field is required'],
            'optionals': ['This field is required'],
            'union': ['This field is required'],
        }, ctx.exception.errors)

        self.assertEqual('<SomeLists {ints=Missing, optionals=Missing, lists=[], models=None, dynamics=Missing}>',
                         repr(model))
        self.assertEqual('<Field ints of type List[int], required>', repr(SomeLists.__fields__[0]))
        self.assertEqual('<Field optionals of type List[str], required>', repr(SomeLists.__fields__[1]))

    def test_basic(self):
        model = SomeLists({
            'ints': [4.0, 5, 6],
            'optionals': ['None', None],
            'lists': [[1], [2, 2.2]],
            'models': [{'field': True}],
            'union': [{'type': 'str'}, {'type': 'bool', 'field': False}],
        })
        self.assertEqual([4, 5, 6], model.ints)
        self.assertIsInstance(model.ints[0], int)
        self.assertEqual(['None', None], model.optionals)
        self.assertEqual([[1], [2, 2.2]], model.lists)
        self.assertIsInstance(model.lists[0][0], float)
        self.assertEqual([MyBoolModel({'field': True})], model.models)
        self.assertEqual([MyStrModel({'field': ''}), MyBoolModel({'field': False})], model.dynamics)
        model.validate()
        self.assertEqual({
            'ints': [4, 5, 6],
            'optionals': ['None', None],
            'lists': [[1.], [2., 2.2]],
            'models': [{'field': True}],
            'union': [{'type': 'str', 'field': ''}, {'type': 'bool', 'field': False}],
        }, model.serialize())
        self.assertEqual('<SomeLists {ints=[(3 items)], optionals=[(2 items)], lists=[(2 items)], models=[(1 items)], '
                         'dynamics=[(2 items)]}>', repr(model))

    def test_inner_validation(self):
        data = {
            'ints': [None, 5],
            'optionals': ['xyz', 'a', None, 'b'],
            'lists': [[1], [None, 2.2]],
            'models': [None, {'bad_field': True}],
            'union': [{'type': 'bool', 'field': False}, {'type': 'str', 'field': '01234'}],
        }
        model = SomeLists(data)
        self.assertEqual([None, 5], model.ints)
        self.assertEqual(['xyz', 'a', None, 'b'], model.optionals)
        self.assertEqual([[1], [None, 2.2]], model.lists)
        self.assertEqual([None, MyBoolModel({'bad_field': True})], model.models)
        self.assertEqual([MyBoolModel({'field': False}), MyStrModel({'field': '01234'})], model.dynamics)
        self.assertEqual(data, model.serialize())
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual({
            'ints': {'0': ['This field is required']},
            'optionals': {'1': ['Must be at least 2 characters long'], '3': ['Must be at least 2 characters long']},
            'lists': {'1': {'0': ['This field is required']}},
            'models': {'0': ['This field is required'],
                       '1': {'bad_field': ['Nope'], 'field': ['This field is required']}},
            'union': {'1': {'field': ['Must be at most 3 characters long']}},
        }, ctx.exception.errors)

    def test_size_validation(self):
        class Sizes(Model):
            min: List[int] = ListField(default=list, min_length=1)
            max: Optional[List[float]] = ListField(default=None, max_length=3)
            exact: List[str] = ListField(min_length=3, max_length=3)
            min_max: List[MyStrModel] = ListField(min_length=1, max_length=3)

        model = Sizes({'max': [1, 2.0, 3.3, 4.4], 'exact': ['a', 'b'], 'min_max': []})
        self.assertEqual({'min': [], 'max': [1, 2.0, 3.3, 4.4], 'exact': ['a', 'b'], 'min_max': []}, model.serialize())
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual({
            'min': ['Provide at least 1 item'],
            'max': ['Provide at most 3 items'],
            'exact': ['Provide exactly 3 items'],
            'min_max': ['Provide 1 to 3 items'],
        }, ctx.exception.errors)
        Sizes({'min': ['4'], 'max': None, 'exact': ['a', 'b', 'c'], 'min_max': [{}, {}]}).validate()
        Sizes({'min': ['4', 2], 'max': {'4.2', 0}, 'exact': ['a', 'a', 'a'], 'min_max': [{}, {}, {}]}).validate()

    def test_field_options(self):
        my_role = Role('my-role')

        class Stuff(Model):
            hide_none: Optional[List[str]] = ListField(hide_none=True)
            hide_empty: Optional[List[str]] = ListField(hide_empty=True, default=None)
            custom_output: List[MyStrModel] = ListField(to_primitive_name='custom')
            no_output: List[int] = ListField(primitive_name='input', to_primitive_name=None)

            @classmethod
            def declare_roles(cls) -> Iterable[RequestedRoleFields]:
                yield my_role.blacklist(cls.custom_output)

        model = Stuff({'hide_none': ['ok'], 'custom': ['nope'], 'custom_output': [{}], 'input': [1, 2]})
        self.assertEqual({'hide_none': ['ok'], 'hide_empty': None, 'custom': [{'field': ''}]}, model.to_primitive())

        model = Stuff({'hide_none': None, 'hide_empty': [], 'custom_output': [{'field': 'abc'}]})
        self.assertEqual({}, model.to_primitive(my_role))
        self.assertEqual({'custom': [{'field': 'abc'}]}, model.to_primitive())

    def test_nested_conversion_errors(self):
        class Collections(Model):
            models: List[SimpleInt]
            ints: List[int]

        with self.assertRaises(ConversionError) as ctx:
            Collections({'models': [{'int': 1}, {'int': 'nan'}, {'int': 'bad'}]})
        self.assertEqual({'models': {'1': {'int': ["Value 'nan' is not an integer number"]}}}, ctx.exception.errors)

        with self.assertRaises(ConversionError) as ctx:
            Collections({'ints': ['bad', 'worse']})
        self.assertEqual({'ints': {'0': ["Value 'bad' is not an integer number"]}}, ctx.exception.errors)

    def test_configuration_error_not_list(self):
        class Bad(Model):
            mismatch: Dict[str, MyStrModel] = ListField(item_field=ModelField())
        with self.assertRaises(ConfigurationError) as ctx:
            Bad()
        self.assertEqual('Field mismatch: ListField cannot be used for annotation typing.Dict[str, '
                         'tests.test_compound_fields.MyStrModel], should use DictField', str(ctx.exception))

    def test_configuration_error_list_item_mismatch(self):
        class Bad(Model):
            worse: List[Optional[bool]] = ListField(item_field=StrField(hide_none=True))
        with self.assertRaises(ConfigurationError) as ctx:
            Bad()
        self.assertEqual('Field worse: StrField cannot be used for annotation bool, should use BoolField',
                         str(ctx.exception))


class MyDicts(Model):
    int_to_int: Dict[int, int]
    str_to_floats: Dict[str, List[float]] = DictField(StrField(min_length=2), default=lambda: {'xy': []})
    bool_to_model: Optional[Dict[bool, MyBoolModel]] = None
    str_to_opt_dict: Dict[str, Optional[Dict[int, int]]] = DictField(value_field=DictField(min_length=1),
                                                                     hide_empty=True)


class TestDictType(TestCase):
    def test_empty(self):
        model = MyDicts()
        self.assertIs(Missing, model.int_to_int)
        self.assertEqual({'xy': []}, model.str_to_floats)
        self.assertIsNone(model.bool_to_model)
        self.assertIs(Missing, model.str_to_opt_dict)
        self.assertEqual({'str_to_floats': {'xy': []}, 'bool_to_model': None}, model.serialize())
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual({
            'int_to_int': ['This field is required'],
            'str_to_opt_dict': ['This field is required'],
        }, ctx.exception.errors)

        self.assertEqual('<MyDicts {int_to_int=Missing, str_to_floats={(1 items)}, bool_to_model=None, '
                         'str_to_opt_dict=Missing}>', repr(model))
        self.assertEqual('<Field int_to_int of type Dict[int, int], required>', repr(MyDicts.__fields__[0]))

        model.str_to_floats = {}
        model.str_to_opt_dict = {}
        self.assertEqual({'bool_to_model': None, 'str_to_floats': {}}, model.serialize())
        self.assertEqual('<MyDicts {int_to_int=Missing, str_to_floats={}, bool_to_model=None, '
                         'str_to_opt_dict={}}>', repr(model))

    def test_basic(self):
        model = MyDicts({
            'int_to_int': {1: 2, 3.0: 4.0},
            'str_to_floats': {},
            'bool_to_model': {True: {'field': False}},
            'str_to_opt_dict': {'x': {'1': 2.0}, 'y': {3.0: '4'}},
        })
        self.assertEqual({1: 2, 3.0: 4.0}, model.int_to_int)
        self.assertEqual({}, model.str_to_floats)
        self.assertEqual({True: MyBoolModel({'field': False})}, model.bool_to_model)
        self.assertEqual({'x': {1: 2}, 'y': {3: 4}}, model.str_to_opt_dict)
        model.validate()
        self.assertEqual({
            'int_to_int': {1: 2, 3.0: 4.0},
            'str_to_floats': {},
            'bool_to_model': {True: {'field': False}},
            'str_to_opt_dict': {'x': {1: 2}, 'y': {3: 4}},
        }, model.serialize())
        self.assertEqual('<MyDicts {int_to_int={(2 items)}, str_to_floats={}, bool_to_model={(1 items)}, '
                         'str_to_opt_dict={(2 items)}}>', repr(model))

    def test_conversion_errors(self):
        data = cast(dict, {
            'int_to_int': {1: 2, 4.2: 4.7},
            'str_to_floats': ['xyz'],
            'bool_to_model': {1: None},
            'str_to_opt_dict': {4: 2},
        })
        with self.assertRaises(ConversionError) as ctx:
            MyDicts(data)
        self.assertEqual({
            'int_to_int': {'4.2': ['Numeric value 4.2 is not an integer']},
        }, ctx.exception.errors)

        data['int_to_int'] = {1: 2, 4: 4.7}
        with self.assertRaises(ConversionError) as ctx:
            MyDicts(data)
        self.assertEqual({
            'int_to_int': {'4': ['Numeric value 4.7 is not an integer']},
        }, ctx.exception.errors)

        data['int_to_int'][4] = 7
        with self.assertRaises(ConversionError) as ctx:
            MyDicts(data)
        self.assertEqual({
            'str_to_floats': ['Expected a dict, got a list'],
        }, ctx.exception.errors)

        data['str_to_floats'] = {'xyz': ['4.2']}
        with self.assertRaises(ConversionError) as ctx:
            MyDicts(data)
        self.assertEqual({
            'bool_to_model': {'1': ['Value must be a boolean or a true/false/yes/no string value']},
        }, ctx.exception.errors)

        data['bool_to_model'] = {True: None}
        with self.assertRaises(ConversionError) as ctx:
            MyDicts(data)
        self.assertEqual({
            'str_to_opt_dict': {'4': ['Expected a dict, got a int']},
        }, ctx.exception.errors)

        data['str_to_opt_dict'][4] = None
        MyDicts(data)

    def test_nested_conversion_errors(self):
        class Collections(Model):
            models: Dict[str, SimpleInt]
            ints: Dict[str, int]

        with self.assertRaises(ConversionError) as ctx:
            Collections({'models': {'x': {'int': 'bad'}, 'y': {'int': 2}, '3': {'int': 'y'}}})
        self.assertEqual({'models': {'x': {'int': ["Value 'bad' is not an integer number"]}}}, ctx.exception.errors)

        with self.assertRaises(ConversionError) as ctx:
            Collections({'ints': {'x': 1, 'y': '2+3', 'z': 'y'}})
        self.assertEqual({'ints': {'y': ["Value '2+3' is not an integer number"]}}, ctx.exception.errors)

    def test_inner_validation(self):
        data = {
            'int_to_int': None,
            'str_to_floats': {'a': [4.7, None, 42]},
            'bool_to_model': {None: {'field': False}, True: {'field': True, 'bad_field': True}, False: None},
            'str_to_opt_dict': {'x': {1: 2}, 'y': {}},
        }
        model = MyDicts(data)
        self.assertIs(None, model.int_to_int)
        self.assertEqual({'a': [4.7, None, 42]}, model.str_to_floats)
        self.assertEqual({
            None: MyBoolModel({'field': False}),
            True: MyBoolModel({'field': True, 'bad_field': True}),
            False: None,
        }, model.bool_to_model)
        self.assertEqual({'x': {1: 2}, 'y': {}}, model.str_to_opt_dict)
        self.assertEqual(data, model.serialize())
        self.assertIsInstance(model.serialize()['str_to_floats']['a'][2], float)
        with self.assertRaises(ValidationError) as ctx:
            model.validate()
        self.assertEqual({
            'int_to_int': ['This field is required'],
            'str_to_floats': {'a': {'1': ['This field is required'],
                                    '_global': ['Must be at least 2 characters long']}},
            'bool_to_model': {'None': ['This field is required'],
                              'True': {'bad_field': ['Nope']},
                              'False': ['This field is required']},
            'str_to_opt_dict': {'y': ['Provide at least 1 item']},
        }, ctx.exception.errors)

    def test_double_required_error(self):
        class RequiredDict(Model):
            dict: Dict[int, int]
        model = RequiredDict({'dict': {4: 2}})
        model.validate()
        self.assertEqual(2, model.dict[4])
        with self.assertRaises(ValidationError) as ctx:
            RequiredDict({'dict': {None: None}}).validate()
        self.assertEqual('dict: None: This field is required', str(ctx.exception))

    def test_configuration_error_not_dict(self):
        class Bad(Model):
            mismatch: List[MyStrModel] = DictField(key_field=BoolField(), value_field=ModelField())
        with self.assertRaises(ConfigurationError) as ctx:
            Bad()
        self.assertEqual('Field mismatch: DictField cannot be used for annotation '
                         'typing.List[tests.test_compound_fields.MyStrModel], should use ListField', str(ctx.exception))

    def test_configuration_error_invalid_key(self):
        class InvalidKey(Model):
            bad: Dict[MyStrModel, bool]
        with self.assertRaises(ConfigurationError) as ctx:
            InvalidKey()
        self.assertEqual('Field bad: DictField keys may only be booleans, numbers or strings: '
                         'typing.Dict[tests.test_compound_fields.MyStrModel, bool]', str(ctx.exception))

    def test_configuration_error_dict_key_mismatch(self):
        class Bad(Model):
            bad: Dict[int, bool] = DictField(key_field=FloatField())
        with self.assertRaises(ConfigurationError) as ctx:
            Bad()
        self.assertEqual('Field bad: FloatField cannot be used for annotation int, should use IntField',
                         str(ctx.exception))

    def test_configuration_error_dict_value_mismatch(self):
        class Bad(Model):
            worse: Dict[int, bool] = DictField(value_field=ModelField())
        with self.assertRaises(ConfigurationError) as ctx:
            Bad()
        self.assertEqual('Field worse: ModelField cannot be used for annotation bool, should use BoolField',
                         str(ctx.exception))



from __future__ import annotations

from typing import List, Optional

from pydantic import BaseModel, validator
from bench.common import execute, benchmark_cpu, memory_bench


class BaseStuff(BaseModel):
    name: str
    flag: bool = False

    @validator('name')
    def name_min_length_2(cls, v):
        if len(v) < 2:
            raise ValueError('must be at least 2 characters long')
        return v


class ListStuff(BaseStuff):
    stuff: Stuff
    list: List[str]

    @validator('list')
    def list_min_length_1(cls, v):
        if len(v) < 1:
            raise ValueError('must have at least 1 items')
        return v


class ModelStuff(BaseModel):
    value: float
    def_value: float = 4.2


class Stuff(BaseStuff):
    optional: Optional[int] = None
    strange: Optional[float] = 4.7
    model: ModelStuff
    items: List[ListStuff]

    @validator('flag', 'optional')
    def flag_set_if_not_optional(cls, flag, values):
        if not flag and values.get('optional') is None:
            raise ValueError('Must be true if optional is not set')
        return flag


ListStuff.update_forward_refs()


if __name__ == '__main__':
    def benchmark(inputs: List[dict], validate: bool):
        for data in inputs:
            if validate:
                model = Stuff(**data)
            else:
                model = Stuff.construct(**data)
            yield model.dict()

    benchmark_cpu(benchmark, depth=4, validate=True)
    # memory_bench(lambda data: Stuff(**data), lambda model: model.dict(), 10000, 4)  # 651 MB



from unittest import TestCase


class TestExamples(TestCase):
    def test_atomic_fields(self):
        from examples.atomic_fields import Employee
        Employee()

    def test_compound_fields(self):
        from examples.compound_fields import Book
        Book()

    def test_model_fields(self):
        from examples.model_fields import Conjunction
        Conjunction()

    def test_schematics_field(self):
        from examples.schematics_field import Branch
        Branch()



from __future__ import annotations
from typing import Any, Optional, TYPE_CHECKING, Union, get_origin, get_args

from stereotype.utils import ConfigurationError

if TYPE_CHECKING:  # pragma: no cover
    from stereotype.fields.base import Field


class AnnotationResolver:
    __slots__ = ['annotation', 'origin', 'optional']

    def __init__(self, annotation: Any):
        """Create helper capable of resolving annotations to Fields, automatically unwrap Optional."""
        self.annotation = annotation
        self.origin = get_origin(annotation)
        self.optional = False
        self._unwrap_optional()

    def _unwrap_optional(self):
        if self.origin is not Union:
            return

        options = get_args(self.annotation)
        non_none = [option for option in options if option not in (type(None),)]
        if len(non_none) == len(options):
            return  # A Union of non-optional types

        self.optional = True
        if len(non_none) == 1:
            self.annotation = non_none[0]
        else:
            # There are more elements in the Union, remove None from it
            self.annotation = Union[tuple(non_none)]
        self.origin = get_origin(self.annotation)

    def __repr__(self):
        return self.annotation.__name__ if hasattr(self.annotation, '__name__') else repr(self.annotation)

    def resolve(self, explict_field: Optional[Field] = None) -> Field:
        if explict_field is None or explict_field is NotImplemented:
            field = self.auto_resolve()
        else:
            field = explict_field.copy_field()  # So that subclasses don't share the same Field and leak validation
        field.init_from_annotation(self)
        field.allow_none = self.optional
        return field

    def auto_resolve(self) -> Field:
        """Attempt to recognize the annotation and create the corresponding Field instance."""
        from stereotype import Model, ListField, DictField, ModelField, DynamicModelField, AnyField
        from stereotype.fields.atomic import ATOMIC_TYPE_MAPPING

        if self.origin is not None:
            if self.origin is list:
                return ListField()
            if self.origin is dict:
                return DictField()
            if self.origin is Union:
                return DynamicModelField()  # Cannot be Optional at this point, taken care of in init
        elif self.annotation is Any:
            return AnyField()

        elif atomic_field := ATOMIC_TYPE_MAPPING.get(self.annotation):
            return atomic_field()

        elif issubclass(self.annotation, Model):
            return ModelField()

        raise ConfigurationError(f'Unrecognized field annotation {self!r} (may need an explicit Field)')

    def incorrect_type(self, field: Field) -> ConfigurationError:
        hint = ''
        try:
            hint = f', should use {type(self.auto_resolve()).__name__}'
        except ConfigurationError:
            pass
        return ConfigurationError(f'{type(field).__name__} cannot be used for annotation {self!r}{hint}')



from __future__ import annotations

from typing import Dict, Any, Iterable, Tuple, get_type_hints, cast, List, Set, Type, TYPE_CHECKING, Optional

from stereotype.fields.annotations import AnnotationResolver
from stereotype.fields.base import Field
from stereotype.fields.serializable import SerializableField
from stereotype.roles import Role, RequestedRoleFields, FinalizedRoleFields
from stereotype.utils import ConfigurationError, Missing

if TYPE_CHECKING:  # pragma: no cover
    from stereotype.model import Model


class ModelMeta(type):
    def __new__(mcs, name: str, bases: Tuple[type, ...], attrs: Dict[str, Any]):
        if not bases:
            return type.__new__(mcs, name, bases, attrs)

        # Only annotated attributes iterated here for the purposes of slots, not serializable fields
        field_names = [name for name, annotation in mcs._iterate_fields(attrs.get('__annotations__', {}))]
        field_values = {}
        for field_name in field_names:
            if field_name in attrs:
                # Field attributes popped as they will become instance attributes instead of class attributes
                field_values[field_name] = attrs.pop(field_name)

        # Find serializable, keep them in `attrs` as they need to remain as properties
        serializable_names = {name for name, _ in mcs._iterate_serializable(attrs)}

        # Using dicts instead of sets to preserve order
        all_slots = {
            **{slot: 0 for parent in bases for slot in parent.__slots__ or getattr(parent, '__abstract_slots__', ())},
            **{slot: 1 for slot in field_names},
            **{slot: 2 for slot in attrs.get('__slots__', ())},
        }
        if attrs.pop('__abstract__', False):
            attrs['__abstract_slots__'] = all_slots
        else:
            attrs['__slots__'] = [name for name in all_slots if name not in serializable_names]
        attrs['__fields__'] = NotImplemented
        attrs['__input_fields__'] = NotImplemented
        attrs['__validated_fields__'] = NotImplemented
        attrs['__role_fields__'] = NotImplemented
        attrs['__roles__'] = NotImplemented

        own_field_names = set(field_names) | serializable_names

        try:
            cls = cast(Type['Model'], type.__new__(mcs, name, bases, attrs))
        except TypeError as e:
            raise ConfigurationError(f'{name}: {e}, if inheriting from multiple models, only one may have __slots__ '
                                     f'(declare abstract models without __slots__ by adding class attribute '
                                     f'`__abstract__ = True`)')
        cls.__initialize_model__ = lambda *_: mcs._initialize_model(cls, bases, own_field_names, field_values)
        return cls

    @classmethod
    def _iterate_fields(mcs, field_annotations: Dict[str, Any]) -> Iterable[Tuple[str, Any]]:
        for name, annotation in field_annotations.items():
            if name.startswith('_'):
                continue
            yield name, annotation

    @classmethod
    def _initialize_model(mcs, cls: Type[Model], bases: Tuple[type, ...],
                          own_field_names: Set[str], field_values: dict):
        from stereotype.model import Model
        model_bases = [base for base in bases if issubclass(base, Model) and base is not Model]
        mcs._ensure_parent_models(model_bases, own_field_names, field_values)

        serializable = list(mcs._analyze_serializable(cls))
        serializable_names = {field.name for field in serializable}
        input_fields = [field for field in mcs._analyze_fields(cls, field_values)
                        if field.name not in serializable_names]
        cls.__input_fields__ = [field.make_input_config() for field in input_fields]
        cls.__validated_fields__ = [field.make_validated_config() for field in input_fields if field.has_validation()]
        cls.__fields__ = input_fields + serializable
        mcs._build_roles(cls, model_bases, own_field_names)

    @classmethod
    def _ensure_parent_models(mcs, model_bases: List[Type[Model]], own_field_names: Set[str], field_values: dict):
        for base in model_bases:
            if base.__fields__ is NotImplemented:
                base.__initialize_model__()
            for field in cast(List[Field], base.__fields__):
                if field.name not in own_field_names:
                    field_values[field.name] = field

    @classmethod
    def _analyze_fields(mcs, cls: Type[Model], field_values: dict) -> Iterable[Field]:
        for name, annotation in mcs._iterate_fields(mcs._resolve_annotations(cls)):
            explicit_field: Optional[Field] = None
            default = Missing
            if name in field_values:
                value = field_values[name]
                if isinstance(value, Field):
                    explicit_field = value
                else:
                    default = value

            try:
                field = AnnotationResolver(annotation).resolve(explict_field=explicit_field)
            except ConfigurationError as e:
                raise ConfigurationError(f'Field {name}: {e}')
            field.init_name(name)
            if default is not Missing:
                field.init_default(default)
            # Note a default could have also been present in the explicit field
            field.check_default()

            validator_method = getattr(cls, f'validate_{name}', None)
            if validator_method is not None:
                field.validator_method = validator_method

            yield field

    @classmethod
    def _iterate_serializable(mcs, attrs: Dict[str, Any]) -> Iterable[Tuple[str, SerializableField]]:
        for name, attr in attrs.items():
            if not isinstance(attr, property) or not hasattr(attr.fget, '__field__'):
                continue
            field = getattr(attr.fget, '__field__')
            if isinstance(field, SerializableField):
                yield name, field

    @classmethod
    def _analyze_serializable(mcs, cls: Type[Model]):
        attrs = {name: getattr(cls, name) for name in dir(cls)}
        for name, field in mcs._iterate_serializable(attrs):
            field.init_name(name)
            yield field

    @staticmethod
    def _resolve_annotations(cls: Type[Model]) -> Dict[str, Any]:
        extra_types: Set[Type[Model]] = cls.resolve_extra_types()
        extra_locals = {typ.__name__: typ for typ in extra_types}
        try:
            return get_type_hints(cls, localns=extra_locals)
        except NameError as e:
            raise ConfigurationError(f'Model {cls.__name__} annotation {str(e)}. If not a global symbol or cannot be '
                                     f'imported globally, use the class method `resolve_extra_types` to provide it.')

    @classmethod
    def _build_roles(mcs, cls: Type[Model], bases: List[Type[Model]], own_field_names: Set[str]):
        roles = mcs._collect_finalized_roles(cls, bases, own_field_names)
        max_role_code = max((role.code for role in roles.keys()), default=0)
        default_role_fields = [field.make_output_config() for field in cls.__fields__]
        cls.__role_fields__ = [default_role_fields] * (max_role_code + 1)
        cls.__roles__ = []

        for role, finalized in roles.items():
            field_configs = [field.make_output_config() for field in cls.__fields__ if field.name in finalized.fields]
            cls.__role_fields__[role.code] = field_configs
            cls.__roles__.append(finalized)

    @classmethod
    def _collect_finalized_roles(mcs, cls: Type[Model], bases: List[Type[Model]], own_field_names: Set[str],
                                 ) -> Dict[Role, FinalizedRoleFields]:
        all_field_names, own_requested_roles = mcs._collect_own_requested_roles(cls)
        all_roles = {finalized.role for base in bases for finalized in base.__roles__} | set(own_requested_roles.keys())
        roles: Dict[Role, FinalizedRoleFields] = {role: FinalizedRoleFields(role) for role in all_roles}

        for base in reversed(bases):
            base_roles: Dict[Role, FinalizedRoleFields] = {finalized.role: finalized for finalized in base.__roles__}
            for role, finalized in roles.items():
                base_finalized = base_roles.get(role)
                if base_finalized is not None:
                    finalized.fields.update(base_finalized.fields)
                elif not role.empty_by_default:
                    finalized.fields.update(field.name for field in base.__fields__)

        for role, finalized in roles.items():
            own_requested = own_requested_roles.get(role)
            if own_requested is not None:
                finalized.update_requested(own_requested, all_field_names, own_field_names)
            elif not role.empty_by_default:
                finalized.fields.update(own_field_names)

        return roles

    @classmethod
    def _collect_own_requested_roles(mcs, cls: Type[Model]) -> Tuple[Set[str], Dict[Role, RequestedRoleFields]]:
        all_field_names = {field.name for field in cls.__fields__}
        own_requested_roles: Dict[Role, RequestedRoleFields] = {}
        for requested in cls.declare_roles():
            if requested.role in own_requested_roles:
                raise ConfigurationError(f'Role {requested.role.name} configured for {cls.__name__} multiple times')
            own_requested_roles[requested.role] = requested
        return all_field_names, own_requested_roles



from __future__ import annotations

# Keep this file synchronized with the `stereotype/compat/README.md` documentation!

from typing import Optional, List

from schematics.models import Model as SchematicsModel
from schematics.types import StringType, IntType

from stereotype import Model, ListField
from stereotype.compat.schematics import SchematicsModelField


class Leaf(SchematicsModel):
    color: str = StringType(required=True, default="green")


class Fruit(SchematicsModel):
    worms: Optional[int] = IntType()


class Branch(Model):
    length: float
    fruit: Optional[Fruit] = SchematicsModelField(hide_none=True, default=None)
    leaves: List[Leaf] = ListField(SchematicsModelField(), max_length=50)


branch = Branch({'length': 42, 'leaves': [{'color': 'yellow'}, {}]})
assert branch.leaves[1].color == 'green'
assert branch.fruit is None
assert branch.serialize() == {'length': 42, 'leaves': [{'color': 'yellow'}, {'color': 'green'}]}

branch.fruit = Fruit({'worms': 2})
assert branch.serialize()['fruit'] == {'worms': 2}
branch.validate()



from random import random
from time import time
from typing import Callable, Any

from stereotype import BoolField


def bench(n: int, executable: Callable[[], Any]):
    started = time()
    for i in range(n):
        executable()
    ended = time()
    print(f'{executable} {(ended - started) / n:.9f}')


_stuff = []


def my_exception_func():
    if random() < 0.001:
        raise ValueError('abc')


def bench_exceptions():
    try:
        my_exception_func()
        return ''
    except ValueError as e:
        return str(e)


def my_iter_error_func():
    if random() < 0.001:
        yield 'abc'


def bench_iter_errors():
    for error in my_iter_error_func():
        return error
    return ''


def my_return_error_func():
    if random() < 0.001:
        return 'abc'
    return None


def bench_return_errors():
    result = my_return_error_func()
    if result is not None:
        return result
    return ''


def bench_getattr_normal():
    (getattr(_stuff, 'append'))


def bench_getattr_internal():
    (_stuff.__getattribute__('append'))


_model = BoolField()
_x = True


def bench_globally_imported_symbol():
    return isinstance(_model, BoolField)


def bench_locally_imported_symbol():
    from stereotype.fields import Field
    return isinstance(_model, Field)


def bench_isinstance_bool():
    global _x
    x = True
    for i in range(1000):
        x = x and isinstance(_x, bool)
    _x = x
    return x


def bench_is_true_false():
    global _x
    x = True
    for i in range(1000):
        x = x and _x is False or _x is True
    _x = x
    return x


def bench_try_inside():
    for i in range(1000):
        try:
            if random()*i > 990:
                raise ValueError('abc')
        except ValueError:
            return i


def bench_try_outside():
    try:
        for i in range(1000):
            if random()*i > 990:
                raise ValueError('abc')
    except ValueError:
        return 999


# bench(1000000, bench_exceptions)
# bench(1000000, bench_iter_errors)
# bench(1000000, bench_return_errors)

# bench(10000000, bench_getattr_normal)
# bench(10000000, bench_getattr_internal)

# bench(10000, bench_isinstance_bool)
# bench(10000, bench_is_true_false)

bench(10000, bench_try_inside)
bench(10000, bench_try_outside)



from __future__ import annotations

from typing import List, Optional

from attrs import define, field, asdict
from bench.common import execute, benchmark_cpu, memory_bench


@define
class BaseStuff:
    name: str = field()
    flag: bool = field(default=False)

    @name.validator
    def name_min_length_2(self, attribute, v):
        if len(v) < 2:
            raise ValueError('must be at least 2 characters long')


@define
class ListStuff(BaseStuff):
    stuff: Stuff = field(converter=lambda data: Stuff(**data), default=None)
    list: List[str] = field(default=list)

    @list.validator
    def list_min_length_1(self, attribute, v):
        if len(v) < 1:
            raise ValueError('must have at least 1 items')


@define
class ModelStuff:
    value: float
    def_value: float = field(default=4.2)


@define
class Stuff(BaseStuff):
    model: ModelStuff = field(converter=lambda data: ModelStuff(**data), default=None)
    items: List[ListStuff] = field(converter=lambda data: [ListStuff(**item) for item in data], default=list)
    optional: Optional[int] = field(default=None)
    strange: Optional[float] = field(default=4.7)

    @optional.validator
    def flag_set_if_not_optional(self, attribute, optional):
        if not self.flag and optional is None:
            raise ValueError('Must be true if optional is not set')


if __name__ == '__main__':
    def benchmark(inputs: List[dict], validate: bool):
        for data in inputs:
            model = Stuff(**data)
            yield asdict(model)

    benchmark_cpu(benchmark, depth=4, validate=True)
    # memory_bench(lambda data: Stuff(**data), lambda model: asdict(model), 10000, 4)  # 339MB



from typing import List

from schematics.exceptions import ValidationError
from schematics.models import Model
from schematics.types import StringType, BooleanType, IntType, FloatType
from schematics.types.compound import ListType, ModelType
from schematics.types.serializable import serializable

from bench.common import execute, benchmark_cpu, memory_bench


class DelayedModelType(ModelType):
    # noinspection PyMissingConstructor
    def __init__(self, model_class_callable, **kwargs):
        super(ModelType, self).__init__(**kwargs)
        super_init = super().__init__
        self.init = lambda: super_init(model_class_callable(), **kwargs)


DelayedStuffModelType = DelayedModelType(lambda: Stuff, required=True)


class BaseStuff(Model):
    name = StringType(required=True, min_length=2)
    flag = BooleanType(required=True, default=False)


class ListStuff(BaseStuff):
    stuff = DelayedStuffModelType
    list = ListType(StringType(required=True), required=True, min_size=1)


class ModelStuff(Model):
    value = FloatType(required=True)
    def_value = FloatType(required=False, default=4.2)


class Stuff(BaseStuff):
    optional = IntType(required=False, default=None)
    strange = FloatType(required=False, default=4.7)
    model = ModelType(ModelStuff, required=True)
    items = ListType(ModelType(ListStuff, required=True), required=True, default=list)

    def validate_flag(self, data, value):
        if not value and data['optional'] is None:
            raise ValidationError('Must be true if optional is not set')


DelayedStuffModelType.init()


# class SerializableTest(Model):
#     stuff = StringType()
#     stuff2 = StringType(serialized_name='stuff3')
#
#     @serializable
#     def other(self):
#         return self.stuff + ' ' + self.stuff
#
#
# print(list(SerializableTest({'stuff': 'abc', 'stuff2': 'abc'}).items()))
# print(SerializableTest({'stuff': 'abc', 'stuff3': 'xyz'}).serialize())


if __name__ == '__main__':
    def benchmark(inputs: List[dict], validate: bool):
        for data in inputs:
            model = Stuff(data)
            if validate:
                model.validate()
            yield model.to_primitive()

    benchmark_cpu(benchmark, depth=4, validate=False)
    # memory_bench(Stuff, lambda model: model.to_primitive(), 10000, 4)  # 1720 MB



# Replacing schematics
#
# We definitely want to get rid of Schematics 1. Schematics 2 isn't much better (actually worse in some cases, like conversion without validation), IMHO the reasonable alternatives are:
#
# * https://github.com/petee-d/stereotype
#   - my library created specifically for this purpose, 3 stars, surprisingly
#   - compared to schematics: ~11x faster conversion, ~45x faster with validation, ~4.93 less memory usage
#   - compatible with how we use schematics
# * https://pydantic-docs.helpmanual.io/
#   - 9k stars on GitHub
#   - compared to stereotype: 1.12x slower conversion, ~2.59x slower with validation, uses 1.86x more memory
#   - IMHO uglier validation, since last year finally can separate validation from conversion
# * https://www.attrs.org/en/stable/
#   - 4k stars on GitHub
#   - compared to stereotype: ~1.27x slower conversion, slightly faster with validation, uses ~ same memory
#   - recursive models are very ugly, has problems with inheritance from models with defaults, cannot convert without validating
#
# Let's discuss the pros and cons in more detail, choose the right replacement and discuss how it could be done.


from __future__ import annotations

import re
from typing import Optional, Union

from stereotype import Model, FloatField


class BaseConstant(Model):
    comment: Optional[str] = None


class BoolConstant(BaseConstant):
    type = 'bool'

    value: bool = False


class StringConstant(BaseConstant):
    type = 'string'

    value: str

    value_regex = re.compile('^[a-zA-Z0-9_]+$')

    def validate_value(self, value: str, context: Optional[dict] = None):
        if not self.value_regex.match(value):
            raise ValueError('Can only contain alphanumeric characters, numbers or underscore, cannot be empty')


class NumberConstant(BaseConstant):
    type = 'number'

    value: float = FloatField(min_value=0.)


class Definition(Model):
    name: str
    attribute: Union[StringConstant, NumberConstant, BoolConstant]


def do_test():
    model = Definition({'name': 'abc', 'attribute': {'type': 'bool', 'value': '4.7'}})
    print(model)
    print(model.serialize())
    model.validate()


if __name__ == '__main__':
    do_test()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="stereotype",
    version="1.1.0",
    author="Peter Dolák",
    author_email="peter@dolak.sk",
    description="Models for conversion and validation of rich data structures.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/petee-d/stereotype",
    packages=['stereotype', 'stereotype.fields', 'stereotype.compat'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)



from copy import deepcopy
from typing import Any, Optional, Type, cast, Iterable

from schematics.exceptions import ValidationError as SchematicsValidationError
from schematics.models import Model as SchematicsModel

from stereotype.fields.annotations import AnnotationResolver
from stereotype.fields.model import ModelField
from stereotype.roles import DEFAULT_ROLE, Role
from stereotype.utils import Missing, ValidationContextType, PathErrorType


class SchematicsModelField(ModelField):
    def __init__(self, *, default: Any = Missing, hide_none: bool = False,
                 primitive_name: Optional[str] = Missing, to_primitive_name: Optional[str] = Missing):
        super().__init__(default=default, hide_none=hide_none,
                         primitive_name=primitive_name, to_primitive_name=to_primitive_name)

        self.type: Type[SchematicsModel] = cast(Type[SchematicsModel], NotImplemented)
        self.native_validate = self.validate

    def init_from_annotation(self, parser: AnnotationResolver):
        if not issubclass(parser.annotation, SchematicsModel):
            raise parser.incorrect_type(self)
        self.type = parser.annotation

    def validate(self, value: SchematicsModel, context: ValidationContextType) -> Iterable[PathErrorType]:
        try:
            value.validate()  # Cannot propagate the context to schematics
        except SchematicsValidationError as e:
            yield from _iterate_validation_errors(e.messages)

    def copy_value(self, value: Any) -> Any:
        if value is None or value is Missing:
            return value
        return deepcopy(value)

    def to_primitive(self, value: Any, role: Role = DEFAULT_ROLE) -> Any:
        if value is None or value is Missing:
            return value
        role_str = role.name if role is not DEFAULT_ROLE else None
        value: SchematicsModel
        return value.to_primitive(role_str)


def _iterate_validation_errors(messages: dict) -> Iterable[PathErrorType]:
    for key, container in messages.items():
        if isinstance(container, list):
            for error in container:
                yield (key,), error
        else:
            for path, error in _iterate_validation_errors(container):
                yield (key,) + path, error
